import 'dart:async';

import 'package:chopper/chopper.dart';
import 'package:collection/collection.dart';
import 'package:libaqueous/src/api/api.dart';
import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/room/model/room_member.dart';
import 'package:libaqueous/src/api/sync/model/invited_room_sync.dart';
import 'package:libaqueous/src/api/sync/model/room_sync.dart';
import 'package:libaqueous/src/api/sync/model/sync_response.dart';
import 'package:libaqueous/src/client/room/room.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/store/store.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';
import 'package:rxdart/rxdart.dart';

mixin ClientEmitter {
  final _roomsSubject = BehaviorSubject<List<Room>>();

  Observable<List<Room>> get roomsObservable => _roomsSubject.stream;
}

class Client with ClientEmitter {
  Client(String homeserver, userID, token, this.store)
      : api = MatrixAPI(homeserver, token),
        id = DateTime.now().millisecondsSinceEpoch,
        userID = userID;

  String get homeServer => api.baseUrl;

  final String userID;

  String get token => api.token;

  final int id;

  int _transactionCount = 0;

  String get transactionID =>
      id.toString() + "q" + (++_transactionCount).toString();

  MatrixAPI api;

  User get myUser => users[userID];

  bool _running = false;

  Store store;

  Map<String, Room> rooms = {};
  Map<String, User> users = {};

  Future<void> loadFromStore() async {
    final roomIDs = await store.roomIDs();
    if (roomIDs?.isEmpty ?? true) return;
    for (final id in roomIDs) {
      final room = await _roomOrCreate(id);
    }

    final userIDs = await store.userIDs();
    if (userIDs?.isEmpty ?? true) return;
    for (final id in userIDs) {
      final user = await _userOrCreate(id);
    }

    _roomsSubject.add(rooms.values.toList());
  }

  void startSync() {
    if (_running) return;
    _running = true;

    _sync();
  }

  void stopSync() {
    _running = false;
  }

  Future<void> logout() async {
    await api.loginAPI.postLogout();
  }

  Future<void> _sync() async {
    final stream = _getEventStream();

    await for (SyncResponse data in stream) {
      await _processSyncResponse(data);
    }
  }

  Stream<SyncResponse> _getEventStream() async* {
    while (_running) {
      final syncToken = await store.syncToken() ?? "";
      print("Previous token: " + syncToken);

      final Response<SyncResponse> resp = await api.syncAPI.sync(
          "{\"room\":{\"state\":{\"lazy_load_members\":true}}}",
          syncToken,
          syncToken.isEmpty,
          30000);
      final syncResp = resp.body;

      yield syncResp;

      await store.setSyncToken(syncResp.nextBatch);
    }
  }

  Future<void> _processSyncResponse(SyncResponse data) async {
    if (data == null) return;

    // TODO: Process and store sync response.
    if (data.toDevice?.events?.isNotEmpty ?? false) {
      for (final e in data.toDevice.events) {}
    }

    if (data.accountData?.events?.isNotEmpty ?? false) {
      for (final e in data.accountData.events) {
        switch (e.type) {
          case "m.direct":
            print(e.content);
            print(e.content.runtimeType);
            (Map<String, List>.from(e.content))
                .forEach((String userID, List roomIDs) async {
              for (final roomID in roomIDs) {
                if (!(roomID is String)) break;
                final room = await _roomOrCreate(roomID as String);

                room.state.isDirect = true;
                room.state.directUserID = userID;

                await store.storeRoom(room);
              }
            });
            break;
        }
      }
    }

    if (data.presence?.events?.isNotEmpty ?? false) {
      for (Event presenceEvent in data.presence.events) {
//        final user = _userOrCreate(presenceEvent.sender);
      }
    }

    final roomsCopy = Map<String, Room>.from(rooms);

    if (data.rooms?.join?.isNotEmpty ?? false) {
      await _processRooms(data.rooms.join, "join");
    }

    if (data.rooms?.leave?.isNotEmpty ?? false) {
      await _processRooms(data.rooms.leave, "leave");
    }

    if (data.rooms?.invite?.isNotEmpty ?? false) {
      await _processInvitedRooms(data.rooms.invite);
    }

    if (!MapEquality().equals(rooms, roomsCopy)) {
      _roomsSubject.add(rooms.values.toList());
    }
  }

  Future<void> _processRooms(
      Map<String, RoomSync> roomSyncMap, String membership) async {
    for (final id in roomSyncMap.keys) {
      final roomSync = roomSyncMap[id];

      int highlightCount = roomSync.unreadNotifications?.highlightCount ?? 0;
      int notificationCount =
          roomSync.unreadNotifications?.notificationCount ?? 0;
      String prevToken = roomSync.timeline?.prevToken ?? "";
      bool limited = roomSync.timeline?.limited ?? false;

      final room = await _roomOrCreate(id)
        ..state.userMembership = membership;

      if (notificationCount != room.state.notificationCount) {
        room.state.notificationCount = notificationCount;
      }

      if (room.timeline.prevToken?.isEmpty ?? true) {
        room.timeline.prevToken = prevToken;
      }

      await room.processStateEvents(roomSync.state?.events);
      await room.processTimelineEvents(roomSync.timeline?.events, false);
      await room.processAccountDataEvents(roomSync.accountData?.events);

      store.storeRoom(room);
    }
  }

  Future<void> _processInvitedRooms(
      Map<String, InvitedRoomSync> roomSyncMap) async {
    for (final id in roomSyncMap.keys) {
      final roomSync = roomSyncMap[id];

      final room = await _roomOrCreate(id);

      room.state.userMembership = RoomMember.MEMBERSHIP_INVITE;
    }
  }

  Future<Room> _roomOrCreate(String id) async {
    if (!rooms.containsKey(id)) {
      final room = Room(
          this,
          await store.roomState(id) ?? RoomState()
            ..roomID = id,
          await store.roomAccountData(id) ?? RoomAccountData(),
          await store.roomTimeline(id) ?? RoomTimeline());
      rooms[id] = room;
      store.storeRoom(room);
    }
    return rooms[id];
  }

  Future<User> _userOrCreate(String id) async {
    if (!users.containsKey(id)) {
      final user = User(
          this,
          await store.userState(id) ?? UserState()
            ..userID = id);
      users[id] = user;
      store.storeUser(user);
    }
    return users[id];
  }
}
