import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/events/model/event_type.dart';
import 'package:libaqueous/src/client/client.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_tag.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';
import 'package:rxdart/rxdart.dart';

mixin RoomEmitter {
  final _timelineSubject = BehaviorSubject<RoomTimeline>();

  Observable<RoomTimeline> get timelineObservable => _timelineSubject.stream;

  final _stateSubject = BehaviorSubject<RoomState>();

  Observable<RoomState> get stateObservable => _stateSubject.stream;
}

class Room with RoomEmitter {
  Room(this.client, this.state, this.accountData, this.timeline) {
    _timelineSubject.add(timeline);
  }

  String get id => state.roomID;

  String get displayName =>
      (state.name ??
          (state.isDirect
              ? client.users[state.directUserID]?.state?.name
              : null)) ??
      state.canonicalAlias ??
      state.roomID;

  Uri get avatarUrl =>
      state.avatarUrl ??
      (state.isDirect
          ? client.users[state.directUserID]?.state?.avatarUrl
          : null);

  bool get isFavourite =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_FAVOURITE) ?? false;

  bool get isLowPriority =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_LOW_PRIORITY) ?? false;

  final RoomState state;
  final RoomAccountData accountData;
  final RoomTimeline timeline;

  final Client client;

  Future<void> processStateEvents(List<Event> events) async {
    if (events?.isEmpty ?? true) return;

    for (final event in events) {
      await processStateEvent(event);
    }
  }

  Future<void> processStateEvent(Event event) async {
    switch (event.type) {
      case EventType.STATE_ROOM_NAME:
        state.name = event.content["name"];
        break;
      case EventType.STATE_ROOM_TOPIC:
        state.topic = event.content["topic"];
        break;
      case EventType.STATE_ROOM_AVATAR:
        state.avatarUrl = Uri.parse(event.content["url"] ?? "");
        break;
      case EventType.STATE_CANONICAL_ALIAS:
        state.canonicalAlias = event.content["alias"];
        break;
      case EventType.STATE_ROOM_MEMBER:
        final avatarUrl = event.content.containsKey("avatar_url") &&
                event.content["avatar_url"] != null
            ? Uri.parse(event.content["avatar_url"])
            : null;

        if (!client.users.containsKey(id)) {
          final user = User(
              this.client,
              UserState()
                ..userID = event.sender
                ..name = event.content["displayname"]
                ..avatarUrl = avatarUrl);
          client.users[event.sender] = user;
          client.store.storeUser(user);
        } else {
          final user = client.users[event.sender];

          user.state
            ..avatarUrl = avatarUrl
            ..name = event.content["displayname"];

          client.store.storeUser(user);
        }
    }
  }

  Future<void> processTimelineEvents(List<Event> events, bool isHistory) async {
    if (events?.isEmpty ?? true) return;

    bool processedStateEvent = false;

    for (final event in events) {
      if (!isHistory && EventType.isStateEvent(event.type)) {
        processedStateEvent = true;
        await processStateEvent(event);
      }
      await processTimelineEvent(event, isHistory);
    }

    if (processedStateEvent) {}

    _timelineSubject.add(timeline);
  }

  Future<void> processTimelineEvent(Event event, bool isHistory) async {
    if (isHistory) {
      timeline.events.insert(0, event);
    } else {
      timeline.events.add(event);
    }
  }

  Future<void> processAccountDataEvents(List<Event> events) async {
    if (events?.isEmpty ?? true) return;

    accountData.tags?.clear();

    for (final event in events) {
      await processAccountDataEvent(event);
    }
  }

  Future<void> processAccountDataEvent(Event event) async {
    if (event.type == "m.tag") {
      if (event.content["tags"] is Map) {
        final tags = (event.content["tags"]) as Map<String, dynamic>;
        if (tags.isEmpty) return;

        tags.forEach((String tag, dynamic data) {
          double order = 0.0;

          if (data is Map<String, dynamic>) {
            final tagData = data as Map<String, dynamic>;
            if (tagData["order"] is int) {
              order = tagData["order"].toDouble();
            } else if (tagData["order"] is double) {
              order = tagData["order"];
            }
          }

          accountData.tags[tag] = order;
        });
      }
    }
  }

  Future<void> postMessage(Map<String, dynamic> json) async {
    await client.api.roomAPI
        .send(this.id, "m.room.message", client.transactionID, json);
  }

  Future<void> postPlainText(String text) async {
    await postMessage({"msgtype": "m.text", "body": text});
  }

  Future<void> postEmote(String text) async {
    await postMessage({"msgtype": "m.emote", "body": text});
  }

  Future<void> getPreviousEvents(int limit) async {
    final resp = (await client.api.roomAPI
            .getMessages(this.id, timeline.prevToken, "", "b", limit, ""))
        .body;
    timeline.prevToken = resp.end;
    if (resp?.chunk?.isEmpty ?? true) return;
    await processTimelineEvents(resp.chunk, true);
  }
}
