import 'dart:convert';
import 'dart:io';

import 'package:libaqueous/src/client/room/room.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/store/store.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';
import 'package:path/path.dart';
import 'package:sembast/sembast.dart' as Sembast;
import 'package:sembast/sembast_io.dart' as SemBastIO;

class SembastStore implements Store {
  final String path;

  Sembast.Database _db;

  Sembast.StoreRef _mainStore = Sembast.StoreRef<String, String>.main();

  Sembast.StoreRef _roomIDStore = Sembast.StoreRef<String, bool>("room_id");
  Sembast.StoreRef _roomStateStore =
  Sembast.StoreRef<String, String>("room_state");
  Sembast.StoreRef _roomTimelineStore =
  Sembast.StoreRef<String, String>("room_timeline");
  Sembast.StoreRef _roomAccountDataStore =
  Sembast.StoreRef<String, String>("room_accountdata");

  Sembast.StoreRef _userIDStore = Sembast.StoreRef<String, bool>("user_id");
  Sembast.StoreRef _userStateStore =
  Sembast.StoreRef<String, String>("user_state");

  SembastStore(this.path);

  @override
  Future<void> open() async {
    String dbPath = join(path, "sembast.db");
    Sembast.DatabaseFactory dbFactory = SemBastIO.databaseFactoryIo;

    _db = await dbFactory.openDatabase(dbPath);
  }

  @override
  Future<void> close() async {
    await _db.close();
  }

  @override
  Future<void> commit() async {
    // TODO: implement commit
  }

  @override
  Future<void> clear() async {
    await close();
    File(join(path, "sembast.db")).delete();
    await open();
  }

  @override
  Future<String> syncToken() async {
    return await _mainStore.record("sync_token").get(_db);
  }

  @override
  Future<void> setSyncToken(String token) async {
    await _mainStore.record("sync_token").put(_db, token);
  }

  @override
  Future<List<String>> roomIDs() async {
    return await _roomIDStore.findKeys(_db, finder: Sembast.Finder());
  }

  @override
  Future<RoomAccountData> roomAccountData(String id) async {
    final rawStr = await _roomAccountDataStore.record(id).get(_db) as String;
    if (rawStr == null || rawStr == "") return null;

    final rawJSON = jsonDecode(rawStr);

    return RoomAccountData.fromJson(rawJSON);
  }

  @override
  Future<RoomState> roomState(String id) async {
    final rawStr = await _roomStateStore.record(id).get(_db) as String;
    if (rawStr == null || rawStr == "") return null;

    final rawJSON = jsonDecode(rawStr);

    return RoomState.fromJson(rawJSON);
  }

  @override
  Future<RoomTimeline> roomTimeline(String id) async {
    final rawStr = await _roomTimelineStore.record(id).get(_db) as String;
    if (rawStr == null || rawStr == "") return null;

    final rawJSON = jsonDecode(rawStr);

    return RoomTimeline.fromJson(rawJSON);
  }

  @override
  Future<void> storeRoom(Room room) async {
    final id = room.id;
    if (id == null || id == "") return;

    await _db.transaction((txn) async {
      if (!await _roomIDStore.record(id).exists(txn)) {
        await _roomIDStore.record(id).put(txn, true);
      }
      await _roomStateStore.record(id).put(txn, room.state.toString());
      await _roomTimelineStore.record(id).put(txn, room.timeline.toString());
      await _roomAccountDataStore
          .record(id)
          .put(txn, room.accountData.toString());
    });
  }

  @override
  Future<List<String>> userIDs() async {
    return await _userIDStore.findKeys(_db, finder: Sembast.Finder());
  }

  @override
  Future<UserState> userState(String id) async {
    final rawStr = await _userStateStore.record(id).get(_db) as String;
    if (rawStr == null || rawStr == "") return null;

    final rawJSON = jsonDecode(rawStr);

    return UserState.fromJson(rawJSON);
  }

  @override
  Future<void> storeUser(User user) async {
    final id = user.id;
    if (id == null || id == "") return;

    await _db.transaction((txn) async {
      if (!await _userIDStore.record(id).exists(txn)) {
        await _userIDStore.record(id).put(txn, true);
      }
      await _userStateStore.record(id).put(txn, user.state.toString());
    });
  }
}
